import { EntityTypes, system, world } from "@minecraft/server";
import { ActionFormData, MessageFormData, ModalFormData } from "@minecraft/server-ui";
import * as UTIL from "./utils";

// protLocationJson = {
//   playername : "",
//   lastloginday: "",
//   prot_location : [
//     { 
//       protname: "",
//       dimension: "",
//       pos_start: ["", "", ""],
//       pos_end: ["", "", ""],
//       shared_flg: false
//     }
//   ],
//   sharedplayer : []
// }
const INITHEIGHT = "x";
const ERRMSG = "§c【エラー】§r";
const UI_OVERWORLD = "textures/blocks/grass_side_carried";
const UI_SHARED_OVERWORLD = "textures/protect/protect_shared_grass_side_carried";
const UI_NETHER = "textures/blocks/netherrack";
const UI_SHARED_NETHER = "textures/protect/protect_shared_netherrack";
const UI_END = "textures/blocks/end_stone";
const UI_SHARED_END = "textures/protect/protect_shared_end_stone";

world.afterEvents.worldInitialize.subscribe((ev) => {
  if(!world.getDynamicProperty(`protLocation_initconfig`)) {
    const initconfig = {
      chestLogFlg : false,
      confExplFlg : 0,
      maxshared : 1,
      deldaysel : 0,
      maxopchunk : 16,
      maxchunk : 4,
      height : 1,
      dimension : 0
    }
    setInitConfig(initconfig);
  }
});

world.afterEvents.playerSpawn.subscribe(ev => {
  if(ev.initialSpawn){
    const player = ev.player;
    initReg(player);
  }
});

world.beforeEvents.playerPlaceBlock.subscribe(ev => {
  const player = ev.player;
  const block = ev.block;
  const face = ev.face;

  const dimension = player.dimension.id;
  const msgFlg = true;
  const explFlg = false;
  const { x, y, z } = compBlockLocation(face, Math.floor(Number(block.x)), Math.floor(Number(block.y)), Math.floor(Number(block.z)));
  if (protectionCheck(player, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.playerBreakBlock.subscribe(ev => {
  const player = ev.player;
  const block = ev.block;

  const dimension = player.dimension.id;
  const msgFlg = true;
  const explFlg = false;
  const { x, y, z } = { x:block.x, y: block.y, z: block.z };
  if (protectionCheck(player, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.itemUse.subscribe(ev => {
  const source = ev.source;

  const dimension = source.dimension.id;
  const msgFlg = false;
  const explFlg = false;
  const { x, y, z } = { x:source.location.x, y: source.location.y, z: source.location.z };
  if (protectionCheck(source, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.itemUseOn.subscribe(ev => {
  const source = ev.source;
  const block = ev.block;
  const face = ev.blockFace;
  
  const dimension = source.dimension.id;
  const msgFlg = false;
  const explFlg = false;
  const { x, y, z } = compBlockLocation(face, Math.floor(Number(block.x)), Math.floor(Number(block.y)), Math.floor(Number(block.z)));

  if (protectionCheck(source, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.playerInteractWithBlock.subscribe(ev => {
  const player = ev.player;
  const block = ev.block;
  const face = ev.blockFace;
  
  const dimension = player.dimension.id;
  const msgFlg = true;
  const explFlg = false;
  const { x, y, z } = compBlockLocation(face, Math.floor(Number(block.x)), Math.floor(Number(block.y)), Math.floor(Number(block.z)));
  if (protectionCheck(player, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.playerInteractWithEntity.subscribe(ev => {
  const player = ev.player;
  const entity = ev.target;

  const dimension = entity.dimension.id;
  const msgFlg = true;
  const explFlg = false;
  const { x, y, z } = { x:entity.location.x, y: entity.location.y, z: entity.location.z };
  if (protectionCheck(player, dimension, x, y, z, msgFlg, explFlg)) {
    ev.cancel = true;
  }
});

world.beforeEvents.explosion.subscribe(ev => {
  const source = ev.source; //dummy
  const dimension = ev.dimension.id;

  const msgFlg = false;
  const explFlg = true;
  const impactedBlocks = ev.getImpactedBlocks();

  for (const block of impactedBlocks) {
    let { x, y, z } = { x:block.x, y: block.y, z: block.z };
    if (protectionCheck(source, dimension, x, y, z, msgFlg, explFlg)) {
      ev.cancel = true;
      break;
    }
  }
});

function protectionCheck(player, dimension, x, y, z, msgFlg, explFlg) {
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const confExplFlg = initconfigJson.confExplFlg;
  const confHeight = initconfigJson.height;

  if (explFlg && confExplFlg == 1) return;
  if (confHeight == 1 && y < 63) return;
  if (confHeight == 2 && y < 0) return;

  const chkChunkLocation = chunkLocation(x, 0, z);
  const chkChunkLocationSplit = chkChunkLocation.split(" ");
  const chkChunkStartPosX = chkChunkLocationSplit[0];
  const chkChunkStartPosZ = chkChunkLocationSplit[1];

  let allProtPlayerList = getAllProtPlayerList();
  if (!allProtPlayerList) return;

  for (let i = 0; i < allProtPlayerList.length; i++) {
    const chkLocationJson = JSON.parse(world.getDynamicProperty(allProtPlayerList[i]));
    if (Object.keys(chkLocationJson.prot_location).length == 0) continue;
    if (!explFlg && chkLocationJson.playername == player.name) continue;

    for (let k = 0; k < Object.keys(chkLocationJson.prot_location).length; k++) {
      let pdimension = chkLocationJson.prot_location[k].dimension;
      let px = chkLocationJson.prot_location[k].pos_start[0];
      let pz = chkLocationJson.prot_location[k].pos_start[2];

      if (pdimension == dimension && chkChunkStartPosX == px && chkChunkStartPosZ == pz) {
        if (!explFlg && Object.keys(chkLocationJson.sharedplayer).length > 0 && chkLocationJson.prot_location[k].shared_flg) {
          for (let i = 0; i < Object.keys(chkLocationJson.sharedplayer).length; i++) {
            if (chkLocationJson.sharedplayer[i] == player.name) {
              return false;
            }
          }
        }
        if (msgFlg) {
          player.sendMessage(`${chkLocationJson.playername} が保護しているチャンクエリアです。${UTIL.getShowStrLocation({x:chkLocationJson.prot_location[k].pos_start[0], y:chkLocationJson.prot_location[k].pos_start[1], z:chkLocationJson.prot_location[k].pos_start[2]})} ~ ${UTIL.getShowStrLocation({x:chkLocationJson.prot_location[k].pos_end[0], y:chkLocationJson.prot_location[k].pos_end[1], z:chkLocationJson.prot_location[k].pos_end[2]})}`);
        }
        return true;
      }
    }
  }
  return false;
}

function compBlockLocation(face, x, y, z) {
  if (face === "Up") y += 1;
  if (face === "Down") y -= 1;
  if (face === "North") z -= 1;
  if (face === "South") z += 1;
  if (face === "East") x += 1;
  if (face === "West") x -= 1;

  return { x:x, y:y, z:z };
};

world.afterEvents.playerInteractWithBlock.subscribe(ev => {
  const player = ev.player;
  const block = ev.block;
  
  if (ev.block.typeId !== "minecraft:chest") return;

  const dimension = player.dimension.id;
  const msgFlg = false;
  const explFlg = false;
  const playerlocation = {
    x: Math.floor(Number(player.location.x)),
    y: Math.floor(Number(player.location.y)),
    z: Math.floor(Number(player.location.z))
  }

  const attentionFlg = protectionCheck(player, dimension, playerlocation.x, playerlocation.y, playerlocation.z, msgFlg, explFlg);

  let chestLog = {
    attentionFlg: attentionFlg,
    date: UTIL.dateFormatNow(UTIL.dateNow(), 'MM/DD hh:mm:ss'),
    username: player.name,
    dimension: player.dimension.id.replace("minecraft:", ""),
    blocklocation: block.location,
    playerlocation: playerlocation
  }
  
  let logsJson = {
    chestLog : []
  }
  
  const logsDymcs = world.getDynamicProperty(`protLocation_log`);
  if (logsDymcs) {
    logsJson = JSON.parse(world.getDynamicProperty(`protLocation_log`));
    if (Object.keys(logsJson.chestLog).length >= 500) {
      logsJson.chestLog.shift();
    }
  }
  logsJson.chestLog.push(chestLog);
  world.setDynamicProperty(`protLocation_log`, JSON.stringify(logsJson));
});

world.beforeEvents.itemUse.subscribe((ev) => {
  const player = ev.source;
  if (ev.itemStack.typeId !== "protect:protective_bell") return;
  if (ev.source.typeId !== EntityTypes.get("minecraft:player").id) return;
  mainForm(player);
//【分析用】///////////////////////////////
// player.sendMessage(`【分析用】動的プロパティの合計バイト数\n${world.getDynamicPropertyTotalByteCount()}`);
//////////////////////////////////////////
});

/********** メイン **********/
function mainForm(player) {
  system.run(() => {
    const mainForm = new ActionFormData();
    mainForm.title(`土地保護`);
    mainForm.body(`土地保護の設定を行います`);
    mainForm.button(`追加`, "textures/ui/accessibility_glyph_color");
    mainForm.button(`一覧と解除`, "textures/ui/book_metatag_default");
    mainForm.button(`共用設定`, "textures/ui/FriendsIcon");
    if (player.isOp() || player.hasTag("op")) {
      mainForm.button(`§t基本設定編集§r\nワールド管理者用`, "textures/ui/gear");
      mainForm.button(`§t全プレイヤー保護一覧§r\nワールド管理者用`, "textures/ui/more-dots");
      mainForm.button(`§tチェストログ§r\nワールド管理者用`, "textures/blocks/chest_front");
      mainForm.button(`§t一括全削除§r\nワールド管理者用`, "textures/ui/cancel");
    }

    mainForm.show(player).then(response => {
      if (response.canceled) return;
      if (response.selection === 0) { addProtect(player) }
      else if (response.selection === 1) { showList(player) }
      else if (response.selection === 2) { setShared(player) }
      else if (response.selection === 3) { showManagement(player) }
      else if (response.selection === 4) { allShow(player) }
      else if (response.selection === 5) { chestLog(player) }
      else if (response.selection === 6) { allDelete(player) };
    });
  });
}
/********** 追加 **********/
function addProtect(player) {
  
  const { x, y, z } = player.location;
  const chunkLocations = chunkLocation(Math.floor(Number(x)), 0, Math.floor(Number(z)));
  const chunkLocationsSplit = chunkLocations.split(" ");
  const chunkStartPosX = chunkLocationsSplit[0];
  const chunkStartPosZ = chunkLocationsSplit[1];
  const chunkEndPosX = chunkLocationsSplit[2];
  const chunkEndPosZ = chunkLocationsSplit[3];

  let protLocationJson = {};
  let protLocationLength = 0;
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const confdeldaysel = initconfigJson.deldaysel;
  const confMaxOpchunk = initconfigJson.maxopchunk;
  const confMaxchunk = initconfigJson.maxchunk;
  const confHeight = initconfigJson.height;
  const confDimension = initconfigJson.dimension;

  if (confDimension == 0) {
    if (player.dimension.id !== "minecraft:overworld") {
      player.sendMessage(`${ERRMSG}オーバーワールドのみ保護可能です。`);
      return;
    }
  }

  initReg(player);

  const allProtPlayerList = getAllProtPlayerList();

  for (let i = 0; i < allProtPlayerList.length; i++) {
    const chkLocationJson = JSON.parse(world.getDynamicProperty(allProtPlayerList[i]));

    for (let k = 0; k < Object.keys(chkLocationJson.prot_location).length; k++) {
      const locationVectorFrom = {
        x: chkLocationJson.prot_location[k].pos_start[0],
        y: chkLocationJson.prot_location[k].pos_start[1],
        z: chkLocationJson.prot_location[k].pos_start[2]
      }
      const locationVectorTo = {
        x: chkLocationJson.prot_location[k].pos_end[0],
        y: chkLocationJson.prot_location[k].pos_end[1],
        z: chkLocationJson.prot_location[k].pos_end[2]
      }

      const chk = checkOverlap(player, chkLocationJson, k, chunkStartPosX, chunkStartPosZ);
      if (chk) {
        if (chkLocationJson.playername == player.name) {
          player.sendMessage(`${ERRMSG}すでに§e自分§rが保護している土地です。${UTIL.getShowStrLocation(locationVectorFrom)} ~ ${UTIL.getShowStrLocation(locationVectorTo)}`);
          return;
        } else {
          player.sendMessage(`${ERRMSG}§e${chkLocationJson.playername}§rが保護している土地です。${UTIL.getShowStrLocation(locationVectorFrom)} ~ ${UTIL.getShowStrLocation(locationVectorTo)}`);
          return;
        }
      }
    }
  }

  protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
  protLocationLength = Object.keys(protLocationJson.prot_location).length;

  let maxChunk = confMaxchunk;
  if (player.isOp() || player.hasTag("op")) maxChunk = confMaxOpchunk;
  
  if (maxChunk <= protLocationLength) {
    player.sendMessage(`${ERRMSG}最大設定可能数§e${maxChunk}§rを超えて保護することができません。現在の土地数:§e${protLocationLength}§r`);
    return;
  }

  system.run(() => {
    const maxSizeProtNameStr = 20;
    const addForm = new ModalFormData();
    addForm.title(`現在地点のチャンクを登録します`);
    addForm.textField(`現在の座標 ${UTIL.getShowStrLocation({x:Math.floor(Number(x)), y:Math.floor(Number(y)), z:Math.floor(Number(z))})}\n地点名 全角§e${maxSizeProtNameStr/2}字§r分まで入力可能です。`, `例:自分の拠点 (任意-入力無しok)`);
    addForm.toggle(`共用設定プレイヤーと共用する`, false);

    addForm.show(player).then(res => {
      if (res.canceled) return;
      const protNameStr = res.formValues[0];
      const sharedFlg = res.formValues[1];

      if(UTIL.bytes(protNameStr) > maxSizeProtNameStr) {
        player.sendMessage(`${ERRMSG}全角§e${maxSizeProtNameStr/2}字分§r(${maxSizeProtNameStr}byte)を超えた地点名を登録できません。入力サイズ${UTIL.bytes(protNameStr)}byte`);
        return;    
      }

      let addLocatData = { 
        protname: protNameStr,
        dimension: player.dimension.id,
        pos_start: [chunkStartPosX, INITHEIGHT, chunkStartPosZ],
        pos_end: [chunkEndPosX, INITHEIGHT, chunkEndPosZ],
        shared_flg : sharedFlg
      }
      protLocationJson.prot_location.push(addLocatData);
  
      world.setDynamicProperty(`protLocation_${player.name}`, JSON.stringify(protLocationJson));

      const locationVectorFrom = {
        x: chunkStartPosX,
        y: INITHEIGHT,
        z: chunkStartPosZ
      }
      const locationVectorTo = {
        x: chunkEndPosX,
        y: INITHEIGHT,
        z: chunkEndPosZ
      }
      player.sendMessage(`${UTIL.getShowStrLocation(locationVectorFrom)} ~ ${UTIL.getShowStrLocation(locationVectorTo)}のチャンクを§a保護§rしました。高さは§e${typeHeightList[confHeight]}`);
      player.sendMessage(`現在の保護数:§e${protLocationLength + 1}§r 保護可能チャンク数:§e${maxChunk}§r`);
      if (confdeldaysel != 0) {
        player.sendMessage(`未ログインが§e${confdeldaysel}§r日続くとその次日0時で解除されます。`);
      }
    })
  })
}
/********** 一覧と解除 **********/
function showList(player){
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const confHeight = initconfigJson.height;
  const confMaxOpchunk = initconfigJson.maxopchunk;
  const confMaxchunk = initconfigJson.maxchunk;

  let maxChunk = confMaxchunk;
  if (player.isOp() || player.hasTag("op")) maxChunk = confMaxOpchunk;

  system.run(() => {
    const listForm = new ActionFormData();
    listForm.title(`保護設定一覧`);

    const playerDymcs = world.getDynamicProperty(`protLocation_${player.name}`);
    let playerDymcsJson;
    if (!playerDymcs) {
      listForm.body(`未登録`);
      listForm.button(`現在は未登録です`, "textures/ui/infobulb");
    } else {
      playerDymcsJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
      if (Object.keys(playerDymcsJson.prot_location).length == 0) {
        listForm.body(`未登録`);
        listForm.button(`現在は未登録です`, "textures/ui/infobulb");
      } else {
        let strChunkCount = `設定可能保護数：§e${maxChunk}§r  登録済み保護数：§e${Object.keys(playerDymcsJson.prot_location).length}§r`;
        listForm.body(`現在の有効高さ座標は§e${typeHeightList[confHeight]}§rです。\nスティーブ&アレックス：共用可\n${strChunkCount}\n保護設定一覧`);

        for (let i = 0; i < Object.keys(playerDymcsJson.prot_location).length; i++) {
          let showStr = getListStr(player, playerDymcsJson, i);
          const icon = setIconDimension(playerDymcsJson.prot_location[i].dimension, playerDymcsJson.prot_location[i].shared_flg);

          listForm.button(showStr, icon);
        }
      }
    }
    listForm.button(`戻る`, "textures/ui/import");

    listForm.show(player).then(res => {
      if (res.canceled) return;
      if (!playerDymcs || Object.keys(playerDymcsJson.prot_location).length == 0) {
        mainForm(player);
        return;
      }
      if (res.selection === Object.keys(playerDymcsJson.prot_location).length) {
        mainForm(player);
        return;
      }
      const allListFlg = false;
      editProtForm(player, playerDymcsJson, res.selection, player.name, allListFlg);
    })
  })
}

/********** 共用設定 **********/
function setShared(player){
  system.run(() => {
    const sharedForm = new ActionFormData();
    sharedForm.title(`共用設定`);

    let bodyStr = "";
    let protLocationJson;
    const playerDymcs = world.getDynamicProperty(`protLocation_${player.name}`);
    if (!playerDymcs) {
      player.sendMessage(`${ERRMSG}土地保護エリア設定が未登録です。`);
      return;
    } else {
      protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
      if (Object.keys(protLocationJson.prot_location).length == 0) {
        player.sendMessage(`${ERRMSG}土地保護エリア設定が未登録です。`);
        return;
      }
    }

    if (Object.keys(protLocationJson.sharedplayer).length == 0) {
      bodyStr = "未登録";
    } else {
      for (let i = 0; i < Object.keys(protLocationJson.sharedplayer).length; i++) {
        bodyStr += protLocationJson.sharedplayer[i] + "\n";
      }
    }
    bodyStr += " ";

    const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
    const confMaxShared = initconfigJson.maxshared;

    sharedForm.body(`現在の設定プレイヤー [最大§e${confMaxShared}§r人まで可能]\n${bodyStr}`);
    sharedForm.button(`追加`, "textures/ui/plus");
    if (Object.keys(protLocationJson.sharedplayer).length > 0) {
      sharedForm.button(`削除`, "textures/ui/cancel");
    }

    sharedForm.show(player).then(response => {
      if (response.canceled) return;
      
      if (response.selection === 0) {
        let protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
        if (protLocationJson.sharedplayer) {
          const sheardPlayerLength = Object.keys(protLocationJson.sharedplayer).length;
          if (confMaxShared < sheardPlayerLength + 1) {
            player.sendMessage(`${ERRMSG}共用プレイヤー最大追加数§e${confMaxShared}§rを超えて追加することができません。現在の設定プレイヤー数:§e${sheardPlayerLength}§r`);
            return;
          }
        }
        addShared(player)
      }
      else if (response.selection === 1) { delShared(player) };
    })
  })
}
function addShared(player) {
  const addSharedForm = new ModalFormData();
  addSharedForm.title(`プレイヤーID登録`);
  addSharedForm.textField(`保護エリアを共用するプレイヤーを登録します\n追加するプレイヤーID`, `例) maikura11`)

  addSharedForm.show(player).then(res => {
    if (res.canceled) return;
    if (!res.formValues[0]) {
      player.sendMessage(`${ERRMSG}プレイヤーIDが入力されていません。`);
      return;
    }
    const addSharedPlayerName = res.formValues[0];

    const chkNameStr = addSharedPlayerName.replace(/\s+/g, "");
    if (addSharedPlayerName.length != chkNameStr.length) {
      player.sendMessage(`${ERRMSG}プレイヤーIDにスペースは入力できません。`);
      return;
    }
    if(!UTIL.isHanEisu(addSharedPlayerName)) {
      player.sendMessage(`${ERRMSG}プレイヤーIDに半角英数以外は入力できません。`);
      return;
    }

    const playerDymcs = world.getDynamicProperty(`protLocation_${player.name}`);
    if (!playerDymcs) {
      initReg(player);
    }
    let protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));

    if (player.name == addSharedPlayerName) {
      player.sendMessage(`${ERRMSG}${player.name} 自分は登録できません。`);
      return;
    }
    if (Object.keys(protLocationJson.sharedplayer).length) {
      for (let i = 0; i < Object.keys(protLocationJson.sharedplayer).length; i++) {
        if (protLocationJson.sharedplayer[i] == addSharedPlayerName) {
          player.sendMessage(`${ERRMSG}${addSharedPlayerName} はすでに共用プレイヤーとして登録中です。`);
          return;
        }
      }
    }

    protLocationJson.sharedplayer.push(addSharedPlayerName);
    world.setDynamicProperty(`protLocation_${player.name}`, JSON.stringify(protLocationJson));
    
    player.sendMessage(`共用設定にプレイヤー ${addSharedPlayerName} を追加しました。\n設定中のプレイヤー\n`);
    for (let i = 0; i < Object.keys(protLocationJson.sharedplayer).length; i++) {
      const sharedplayer = protLocationJson.sharedplayer[i];
      player.sendMessage(sharedplayer);
    }
  })
}

function delShared(player) {
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const confMaxshared = initconfigJson.maxshared;
  
  let protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
  let sharedplayerLength = Object.keys(protLocationJson.sharedplayer).length;

  system.run(() => {
    const listSharedForm = new ActionFormData();
    listSharedForm.title(`共用設定一覧`);
    listSharedForm.body(`現在の設定可能プレイヤー数の最大は§e${confMaxshared}§rです。\n登録済みの共用設定一覧[§e${sharedplayerLength}§r]`);

    for (let i = 0; i < sharedplayerLength; i++) {
      listSharedForm.button(protLocationJson.sharedplayer[i]);
    }

    listSharedForm.button(`戻る`, "textures/ui/import");

    listSharedForm.show(player).then(res => {
      if (res.canceled) return;
      if (res.selection === Object.keys(protLocationJson.sharedplayer).length) {
        setShared(player);
        return;
      }

      let delId = protLocationJson.sharedplayer[res.selection];
      const delSharedForm = new MessageFormData()
      .title(`共用解除`)
      .body(`§c§l【注意！】§r選択した「${delId}」の共用設定を解除してよろしいですか？`)
      .button1(`解除しない`)
      .button2(`解除！`);
    
      delSharedForm.show(player).then(response => {
        if (response.canceled) return;
        if (response.selection === 0) {
          delShared(player);
          return;
        }
  
        if (response.selection === 1) {
          protLocationJson.sharedplayer.splice(res.selection, 1);
          world.setDynamicProperty(`protLocation_${player.name}`, JSON.stringify(protLocationJson));
          player.sendMessage(`${delId} を解除しました。`);
        }
      })
    })
  })
}

/********** 管理 **********/
const typeHeightList = ["Y:無制限", "Y:63以上に有効", "Y:0以上に有効"];
const typeDimList = ["オーバーワールドのみ", "全てのディメンション"];
const confExplList  = ["爆発による破壊不可", "爆発による破壊を許可"];

function showManagement(player){
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));

  system.run(() => {
    const setForm = new ModalFormData()
      .title(`土地保護の設定管理`)
      .toggle('チェストログを記録 ※OFF時ログ削除', initconfigJson.chestLogFlg)
      .dropdown("爆発によるブロック破壊", confExplList, initconfigJson.confExplFlg)
      .slider("共用プレイヤー最大追加数", 0, 63, 1, initconfigJson.maxshared)
      .slider("自動削除する未ログイン経過日数\n例)0:削除なし 1:毎日ログインで削除なし\n", 0, 60, 1, initconfigJson.deldaysel)
      .slider("管理者の設定可能な最大チャンク数", 1, 25, 1, initconfigJson.maxopchunk)
      .slider("メンバーの設定可能な最大チャンク数", 1, 9, 1, initconfigJson.maxchunk)
      .dropdown("高さの有効座標", typeHeightList, initconfigJson.height)
      .dropdown("有効ディメンション", typeDimList, initconfigJson.dimension);

      setForm.show(player).then(res => {
        if (res.canceled) return;
        const resChestLogFlg = res.formValues[0];
        const resConfExplFlg = res.formValues[1];
        const resMaxshared = res.formValues[2];
        const resDelDaysEl = res.formValues[3];
        const resOpChunk = res.formValues[4];
        const resChunk = res.formValues[5];
        const resHeight = res.formValues[6];
        const resDimension = res.formValues[7];

        const initconfig = {
          chestLogFlg : resChestLogFlg,
          confExplFlg : resConfExplFlg,
          maxshared : resMaxshared,
          deldaysel : resDelDaysEl,
          maxopchunk : resOpChunk,
          maxchunk : resChunk,
          height : resHeight,
          dimension : resDimension
        }
        setInitConfig(initconfig);
        player.sendMessage(`土地保護設定が完了しました。現在は下記の通りです。`);
        player.sendMessage(`チェストログを記録: §e${resChestLogFlg}`);
        player.sendMessage(`爆発によるブロック破壊: §e${confExplList[resConfExplFlg]}`);
        player.sendMessage(`共用プレイヤー最大追加数: §e${resMaxshared}`);
        player.sendMessage(`自動削除する経過日数: §e${resDelDaysEl}`);
        player.sendMessage(`管理者の設定可能な最大チャンク数: §e${resOpChunk}`);
        player.sendMessage(`メンバーの設定可能な最大チャンク数: §e${resChunk}`);
        player.sendMessage(`高さの有効座標: §e${typeHeightList[resHeight]}`);
        player.sendMessage(`有効ディメンション: §e${typeDimList[resDimension]}`);

        if (!resChestLogFlg) {
          world.setDynamicProperty(`protLocation_log`, );
        }
      })
  });
}

/********** 全一覧 **********/
function allShow(player){
  system.run(() => {
    const allListForm = new ActionFormData();
    allListForm.title(`全プレイヤー保護一覧`);

    const allProtPlayerList = getAllProtPlayerList();

    let returnBtnCount = 0;
    let showListSort = [];
    if (!allProtPlayerList) {
      returnBtnCount += 1;
      allListForm.body(`未登録`);
      allListForm.button(`現在は未登録です`, "textures/ui/infobulb");
    } else {
      let protFlg = false;
      for (let i = 0; i < allProtPlayerList.length; i++) {
        const playerProt = JSON.parse(world.getDynamicProperty(allProtPlayerList[i]));
        if (Object.keys(playerProt.prot_location).length >= 1) {
          protFlg = true;
          break;
        }
      }
      if (!protFlg) {
        returnBtnCount += 1;
        allListForm.body(`未登録`);
        allListForm.button(`現在は未登録です`, "textures/ui/infobulb");
      } else {
        allListForm.body(`登録済みの保護設定一覧（ログイン日順）`);

        for (let i = 0; i < allProtPlayerList.length; i++) {
          const playerLocationJson = JSON.parse(world.getDynamicProperty(allProtPlayerList[i]));
          let dtlProtNum = Object.keys(playerLocationJson.prot_location).length;
          if (dtlProtNum >= 1) {
            const nameStr = playerLocationJson.playername;
            const lastLoginDate = UTIL.getStringToDate(playerLocationJson.lastloginday);
            const dayStr = UTIL.dateFormatNow(lastLoginDate, 'MM/DD');
            
            const addList = {nameStr, dtlProtNum, dayStr, lastLoginDate};
            showListSort.push(addList);

            returnBtnCount += 1;
          }
        }

        showListSort.sort((a, b) => Date.parse(b.lastLoginDate) - Date.parse(a.lastLoginDate));

        for (let i = 0; i < showListSort.length; i++) {
          allListForm.button(`${showListSort[i].nameStr} 保護数：${showListSort[i].dtlProtNum}\n最終ログイン日：${showListSort[i].dayStr}`, "textures/ui/chevron_right");
        }
      }
      allListForm.button(`戻る`, "textures/ui/import");

      allListForm.show(player).then(res => {
        if (res.canceled) return;
        if (!allProtPlayerList || !protFlg) {
          mainForm(player);
          return;
        }
        if (res.selection === returnBtnCount) {
          mainForm(player);
          return;
        }
        const tgtPlayerName = showListSort[res.selection].nameStr;
        const pProtListForm = new ActionFormData();
        pProtListForm.title(`保護設定一覧`);
        pProtListForm.body(`【${tgtPlayerName}】の\n登録済みの保護設定一覧`);

        const showPlayerJson = JSON.parse(world.getDynamicProperty(`protLocation_${tgtPlayerName}`));
        for (let i = 0; i < Object.keys(showPlayerJson.prot_location).length; i++) {
          let showStr = getListStr(player, showPlayerJson, i);
          const icon = setIconDimension(showPlayerJson.prot_location[i].dimension, showPlayerJson.prot_location[i].shared_flg);
          pProtListForm.button(showStr, icon);
        }
        pProtListForm.button(`戻る`, "textures/ui/import");

        pProtListForm.show(player).then(res => {
          if (res.canceled) return;
          if (res.selection === Object.keys(showPlayerJson.prot_location).length) {
            allShow(player);
            return;
          }
          const allListFlg = true;
          deletePlotForm(player, showPlayerJson, res.selection, tgtPlayerName, allListFlg);
        })
      })
    }
  })
}
/********** チェストログ **********/
function chestLog(player){
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const chestLogFlg = initconfigJson.chestLogFlg;
  if (!chestLogFlg) {
    player.sendMessage(`${ERRMSG}チェストログの記録がOFFです。基本設定で変更可能です。`);
    return;
  }

  const chestLogDymcs = world.getDynamicProperty(`protLocation_log`);
  if (!chestLogDymcs) {
    player.sendMessage(`${ERRMSG}チェストを開けたログがありません。`);
    return;
  }

  system.run(() => {
    const chestLogForm = new ActionFormData();
    chestLogForm.title(`チェストログ`);

    let logsJson = JSON.parse(world.getDynamicProperty(`protLocation_log`));
    let chestLogStr = "";
    for (let i = 0; i < Object.keys(logsJson.chestLog).length; i++) {
      if (logsJson.chestLog[i].attentionFlg) {
        chestLogStr += "§e§l";
      }
      chestLogStr += "<" + logsJson.chestLog[i].date + "|" + logsJson.chestLog[i].username + "|" + logsJson.chestLog[i].dimension + "§r"
       + "\nchest>" + UTIL.getShowStrLocation(logsJson.chestLog[i].blocklocation)
       + "player>" + UTIL.getShowStrLocation(logsJson.chestLog[i].playerlocation) + ">\n";
    }

    chestLogForm.body(`${chestLogStr}`);
    chestLogForm.button(`戻る`, "textures/ui/import");
    chestLogForm.button(`閉じる`, "textures/ui/x_dropdown_default");

    chestLogForm.show(player).then(response => {
      if (response.canceled) return;
      if (response.selection === 0) { mainForm(player) }
      else if (response.selection === 1) return;
    })
  })
}
/********** 全削除 **********/
function allDelete(player){
  const reserveAllDeleteForm = new MessageFormData()
  .title(`全削除`)
  .body(`§c§l【注意！】§rワールドすべての保護設定を削除してよろしいですか。削除後は戻せません。`)
  .button1(`削除しない`)
  .button2(`全削除！`);

  reserveAllDeleteForm.show(player).then(response => {
    if (response.canceled) return;
    if (response.selection === 0) {
      mainForm(player);
      return;
    } else if (response.selection === 1) {
      const allDeleteForm = new MessageFormData()
      .title(`全削除`)
      .body(`§c§l【注意！】§r本当に削除してよろしいですか。削除後は戻せません。`)
      .button1(`削除しない`)
      .button2(`全削除！`);
    
      allDeleteForm.show(player).then(response => {
        if (response.canceled) return;
        if (response.selection === 0) {
          mainForm(player);
          return;
        } else if (response.selection === 1) {
          const tmpCconfig = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
          world.clearDynamicProperties();
          player.sendMessage(`ワールドすべての保護設定を削除しました。`);
          world.setDynamicProperty(`protLocation_initconfig`, JSON.stringify(tmpCconfig));
        }
      })
    }
  })
}

/********** 自動削除 **********/
system.runInterval(() => {
  const initconfigJson = JSON.parse(world.getDynamicProperty(`protLocation_initconfig`));
  const confDelDaysEl = initconfigJson.deldaysel;

  if (confDelDaysEl == 0) return;

  const nowDate = UTIL.dateNow();

  const allProtPlayerList = getAllProtPlayerList();
  for (let i = 0; i < allProtPlayerList.length; i++) {
    const delLocationJson = JSON.parse(world.getDynamicProperty(allProtPlayerList[i]));
    const lastLoginDate = UTIL.getStringToDate(delLocationJson.lastloginday);

    let diff = nowDate.getTime() - lastLoginDate.getTime();
    let days = Math.floor(diff / (1000 * 60 * 60 * 24));

    if (confDelDaysEl < days) {
      world.setDynamicProperty(allProtPlayerList[i],);
    };
  }

}, 72000); //60m

system.runInterval(() => {
  const nowDate = UTIL.dateNow();
  if(nowDate.getHours() == 0 && nowDate.getMinutes() == 0) {
    const Players = [...world.getPlayers()];

    if(Players.length) {
      Players.forEach(joinPlayer => {
        let protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${joinPlayer.name}`));
        protLocationJson.lastloginday = UTIL.getDateToString(nowDate);
        world.setDynamicProperty(`protLocation_${joinPlayer.name}`, JSON.stringify(protLocationJson));
      })
    }
  }
}, 20);

function editProtForm(player, json, selectres, id, allListFlg){
  let selectStr = getListStr(player, json, selectres);
  const editForm = new MessageFormData()
  .title(`設定編集`)
  .body(`「${selectStr}」\n\n共用設定の更新または\n保護設定を解除しますか？`)
  .button1(`保護を解除`)
  .button2(`共用設定`);

  editForm.show(player).then(response => {
    if (response.canceled) return;
    if (response.selection === 0) { deletePlotForm(player, json, selectres, id, allListFlg) }
    else if (response.selection === 1) { editProtSharedForm(player, json, selectres, id) }
  })
}

const editSharedList  = ["共用不可", "この土地を共用可"];
function editProtSharedForm(player, json, selectres, id){
  let selectStr = getListStr(player, json, selectres);
  let sharedInit = Number(json.prot_location[selectres].shared_flg);

  const editProtSharedForm = new ModalFormData()
  .title(`現在地点のチャンクを登録します`)
  .dropdown(`選択した「${selectStr}」の共用設定を更新`, editSharedList, sharedInit);

  editProtSharedForm.show(player).then(res => {
    if (res.canceled) return;
    const sharedRes = res.formValues[0];

    let sharedFlg = Boolean(sharedRes);
    json.prot_location[selectres].shared_flg = sharedFlg;

    world.setDynamicProperty(`protLocation_${id}`, JSON.stringify(json));
    player.sendMessage(`${selectStr} の共用設定を更新しました。§e${editSharedList[sharedRes]}§r`);
  })
}

function deletePlotForm(player, json, selectres, id, allListFlg){
  let delNo = selectres;
  let selectStr = getListStr(player, json, delNo);

  const selectForm = new MessageFormData()
  .title(`設定解除`)
  .body(`§c§l【注意！】§r選択した「${selectStr}」の保護設定を解除してよろしいですか？`)
  .button1(`解除しない`)
  .button2(`解除！`);

  selectForm.show(player).then(response => {
    if (response.canceled) return;
    if (response.selection === 0 && !allListFlg) {
      showList(player);
      return;
    }
    if (response.selection === 0 && allListFlg) {
      allShow(player);
      return;
    }
    if (response.selection === 1) {
      json.prot_location.splice(delNo, 1);
      world.setDynamicProperty(`protLocation_${id}`, JSON.stringify(json));
      player.sendMessage(`${selectStr} を解除しました。`);
    }
  })
}

function getListStr(player, json, i) {
  let showStr;
  if(json.prot_location[i].protname) {
    showStr = json.prot_location[i].protname + "§r";
  } else {
    showStr = "[名称未定]§r";
  };
  const { x, y, z } = player.location;  
  const chunkLocations = chunkLocation(Math.floor(Number(x)), 0, Math.floor(Number(z)));
  const chunkLocationsSplit = chunkLocations.split(" ");
  const chunkStartPosX = chunkLocationsSplit[0];
  const chunkStartPosZ = chunkLocationsSplit[1];

  const chk = checkOverlap(player, json, i, chunkStartPosX, chunkStartPosZ);
  if(chk) showStr += " §m現在地点§r";

  const locationVectorFrom = {
    x: json.prot_location[i].pos_start[0],
    y: json.prot_location[i].pos_start[1],
    z: json.prot_location[i].pos_start[2]
  }
  const locationVectorTo = {
    x: json.prot_location[i].pos_end[0],
    y: json.prot_location[i].pos_end[1],
    z: json.prot_location[i].pos_end[2]
  }

  showStr += `\n${UTIL.getShowStrLocation(locationVectorFrom)} ~ ${UTIL.getShowStrLocation(locationVectorTo)}`;

  return showStr;
}

function setIconDimension(dimension, sharedFlg) {
  if (dimension == `minecraft:overworld` && !sharedFlg) { return UI_OVERWORLD; }
  else if (dimension == `minecraft:overworld` && sharedFlg) { return UI_SHARED_OVERWORLD; }
  else if (dimension == `minecraft:nether` && !sharedFlg) { return UI_NETHER; }
  else if (dimension == `minecraft:nether` && sharedFlg) { return UI_SHARED_NETHER; }
  else if (dimension == `minecraft:the_end` && !sharedFlg) { return UI_END; }
  else if (dimension == `minecraft:the_end` && sharedFlg) { return UI_SHARED_END; }
}

function setInitConfig(json) {
  world.setDynamicProperty(`protLocation_initconfig`, JSON.stringify(json));
}

function initReg(player) {
  const nowDateStr = UTIL.getDateToString(UTIL.dateNow());
  let protLocationJson;

  const playerDymcs = world.getDynamicProperty(`protLocation_${player.name}`);
  if (!playerDymcs) {
    protLocationJson = {
      playername : player.name,
      lastloginday: nowDateStr,
      prot_location : [],
      sharedplayer : [],
      shared_flg : false
    }
  } else {
    protLocationJson = JSON.parse(world.getDynamicProperty(`protLocation_${player.name}`));
    protLocationJson.lastloginday = nowDateStr;
  }
  world.setDynamicProperty(`protLocation_${player.name}`, JSON.stringify(protLocationJson));
}

function getAllProtPlayerList() {
  const dymcsList = world.getDynamicPropertyIds();
  const pExceptList = ['protLocation_initconfig', 'protLocation_log'];
  let allProtPlayerList = dymcsList.filter(data => !pExceptList.includes(data));

  return allProtPlayerList;
}

 function checkOverlap(player, json, i, cx, xz) {
  const chkStr = json.prot_location[i].dimension + json.prot_location[i].pos_start;
  const addStr = player.dimension.id + cx + "," + INITHEIGHT + "," + xz;
  if (chkStr == addStr) {
    return true;
  }
  return false;
}

function chunkLocation(x, y, z) {
  const chunkvalue = 16;

  var rx = x % chunkvalue;
  var rz = z % chunkvalue;
  var ps1_l = (x - rx) + ((x < 0 && rx != 0) ? -16 : 0); //x
  var ps1_r = (z - rz) + ((z < 0 && rz != 0) ? -16 : 0); //y
  var ps4_l = ps1_l + 15; //x
  var ps4_r = ps1_r + 15; //y

  return ps1_l + " " + ps1_r + " " + ps4_l + " " + ps4_r;
}
